// Copyright (c) 2020, 2021 Famedly
//
// This software is released under the MIT License.
// https://opensource.org/licenses/MIT

import 'package:flutter/cupertino.dart';
import 'package:flutter/material.dart';
import 'package:provider/provider.dart';
import 'view_data.dart';

class AdaptivePageLayout extends StatefulWidget {
  final String initialRoute;
  final double columnWidth;
  final Widget Function(BuildContext) topBarBuilder;
  final ViewData Function(RouteSettings) onGenerateRoute;
  final Route Function(
          Widget Function(BuildContext) builder, RouteSettings routeSettings)
      routeBuilder;
  final Color dividerColor;
  final double dividerWidth;
  final bool safeAreaOnColumnView;

  static AdaptivePageLayoutState of(BuildContext context) =>
      Provider.of<AdaptivePageLayoutState>(context, listen: false);

  /// Use this widget as the [home] of your [MaterialApp] for example and as a
  /// replacement of the [Navigator].
  ///
  /// It will use a default [Navigator] on less wide screens like mobile phones
  /// and a 2 or 3 column layout on wide screens.
  ///
  /// You have to define [onGenerateRoute] to
  /// set the existing routes of your app. Set [initialRoute] to set the initial
  /// route and [columnWidth] to change the maximum width of the left and right
  /// columns. [topBar] provides a way to display a custom topBar which you can
  /// display or hide using the [ViewData]. [routeBuilder] lets your control
  /// which RouteBuilder is used when the app is in 1-column mode.
  const AdaptivePageLayout({
    Key key,
    this.initialRoute = '/',
    this.columnWidth = 360.0,
    this.topBarBuilder,
    @required this.onGenerateRoute,
    this.routeBuilder,
    this.dividerColor = Colors.black,
    this.dividerWidth = 1.0,
    this.safeAreaOnColumnView = true,
  }) : super(key: key);
  @override
  AdaptivePageLayoutState createState() => AdaptivePageLayoutState();
}

class AdaptivePageLayoutState extends State<AdaptivePageLayout> {
  final GlobalKey<NavigatorState> _navigator = GlobalKey<NavigatorState>();
  List<ViewData> viewDataStack;
  ViewData get currentViewData => viewDataStack.last;

  @override
  void initState() {
    viewDataStack = [
      widget.onGenerateRoute(RouteSettings(name: widget.initialRoute))
    ];
    super.initState();
  }

  bool canPop() =>
      viewDataStack.length > 1 || _navigator.currentState?.canPop() == true;

  Future<void> pushNamed(String routeName, {Object arguments}) {
    if (columnMode(context)) {
      setState(() {
        viewDataStack.add(
          widget.onGenerateRoute(
            RouteSettings(name: routeName, arguments: arguments),
          ),
        );
      });
      return null;
    }
    return _navigator.currentState.pushNamed(routeName, arguments: arguments);
  }

  Future<void> popAndPushNamed(String routeName, {Object arguments}) {
    if (columnMode(context)) {
      if (viewDataStack.isNotEmpty) {
        setState(() {
          viewDataStack.removeLast();
          viewDataStack.add(
            widget.onGenerateRoute(
              RouteSettings(name: routeName, arguments: arguments),
            ),
          );
        });
      }
      return null;
    }
    return _navigator.currentState
        .popAndPushNamed(routeName, arguments: arguments);
  }

  void pop() {
    if (columnMode(context)) {
      if (viewDataStack.isNotEmpty) {
        setState(() {
          viewDataStack.removeLast();
        });
      }
      return;
    }
    return _navigator.currentState.pop();
  }

  void popUntilIsFirst() {
    if (columnMode(context)) {
      setState(() {
        viewDataStack.removeRange(1, viewDataStack.length);
      });
      return;
    }
    return _navigator.currentState.popUntil((r) => r.isFirst);
  }

  Future<T> pushNamedAndRemoveUntilIsFirst<T>(String routeName,
      {Object arguments}) {
    if (columnMode(context)) {
      setState(() {
        viewDataStack.removeRange(1, viewDataStack.length);
        viewDataStack.add(
          widget.onGenerateRoute(
            RouteSettings(name: routeName, arguments: arguments),
          ),
        );
      });
      return null;
    }
    return _navigator.currentState.pushNamedAndRemoveUntil(
      routeName,
      (r) => r.isFirst,
      arguments: arguments,
    );
  }

  Future<T> pushNamedAndRemoveAllOthers<T>(String routeName,
      {Object arguments}) {
    if (columnMode(context)) {
      setState(() {
        viewDataStack.clear();
        viewDataStack.add(
          widget.onGenerateRoute(
            RouteSettings(name: routeName, arguments: arguments),
          ),
        );
      });
      return null;
    }
    return _navigator.currentState.pushNamedAndRemoveUntil(
      routeName,
      (r) => false,
      arguments: arguments,
    );
  }

  bool columnMode(BuildContext context) =>
      MediaQuery.of(context).size.width > 2.5 * widget.columnWidth;

  bool threeColumnMode(BuildContext context) =>
      MediaQuery.of(context).size.width > 3.5 * widget.columnWidth;

  @override
  Widget build(BuildContext context) {
    return Provider(
      create: (_) => this,
      builder: (context, _) => WillPopScope(
        onWillPop: () {
          if (canPop()) {
            pop();
            return Future.value(false);
          }
          return Future.value(true);
        },
        child: columnMode(context)
            ? SafeArea(
                top: widget.safeAreaOnColumnView,
                left: widget.safeAreaOnColumnView,
                right: widget.safeAreaOnColumnView,
                bottom: widget.safeAreaOnColumnView,
                child: Container(
                  color: Theme.of(context).backgroundColor,
                  child: Column(
                    children: [
                      if (currentViewData.displayTopBar &&
                          widget.topBarBuilder != null) ...{
                        widget.topBarBuilder(context),
                        Container(
                          height: widget.dividerWidth,
                          color: widget.dividerColor,
                        ),
                      },
                      Expanded(
                        child: Row(
                          children: [
                            if (currentViewData.leftView != null ||
                                currentViewData.emptyView != null) ...{
                              Container(
                                clipBehavior: Clip.antiAlias,
                                decoration: BoxDecoration(),
                                width: widget.columnWidth,
                                child:
                                    currentViewData.leftView?.call(context) ??
                                        currentViewData.mainView(context),
                              ),
                              Container(
                                width: widget.dividerWidth,
                                color: widget.dividerColor,
                              ),
                            },
                            if (currentViewData.rightView == null ||
                                threeColumnMode(context))
                              if (currentViewData.expandRightView)
                                ClipRRect(
                                  child: currentViewData.emptyView
                                          ?.call(context) ??
                                      currentViewData.mainView(context),
                                )
                              else
                                Expanded(
                                  child: ClipRRect(
                                    child: currentViewData.emptyView
                                            ?.call(context) ??
                                        currentViewData.mainView(context),
                                  ),
                                ),
                            if (currentViewData.rightView != null) ...{
                              Container(
                                width: widget.dividerWidth,
                                color: widget.dividerColor,
                              ),
                              if (!threeColumnMode(context) ||
                                  currentViewData.expandRightView) ...{
                                Expanded(
                                  child: ClipRRect(
                                      child:
                                          currentViewData.rightView(context)),
                                ),
                              } else
                                Container(
                                  clipBehavior: Clip.antiAlias,
                                  decoration: BoxDecoration(),
                                  width: widget.columnWidth,
                                  child: currentViewData.rightView(context),
                                ),
                            },
                          ],
                        ),
                      ),
                    ],
                  ),
                ),
              )
            : Navigator(
                key: _navigator,
                initialRoute: widget.initialRoute,
                onGenerateRoute: (settings) {
                  if (widget.routeBuilder != null) {
                    return widget.routeBuilder(
                      (context) =>
                          widget.onGenerateRoute(settings).focusView(context),
                      settings,
                    );
                  }
                  return MaterialPageRoute(
                    builder: (context) =>
                        widget.onGenerateRoute(settings).focusView(context),
                  );
                },
              ),
      ),
    );
  }
}
